# automatically generated by the FlatBuffers compiler, do not modify

# namespace: events

class Method(object):
    other = 0
    get = 1
    head = 2
    options = 3
    post = 4
    put = 5
    patch = 6
    delete = 7

